@@ -1,3 +1,9 @@
+#!/usr/bin/env python
+# coding: utf-8
+
+# Copyright (c) Saga Inc.
+# Distributed under the terms of the GPL License.
+
 """
 For each column header, Mito creates a static `column_id` that
 is used to identify that column header throughout the entire course
@@ -12,13 +18,15 @@ of the column header. For string column headers, they are just
 themselves.
 """
 import random
-from typing import Any, Dict, List
+from typing import Any, Collection, Dict, List
+
 import pandas as pd
 
 from mitosheet.errors import make_no_column_error
+from mitosheet.types import ColumnHeader, ColumnID, MultiLevelColumnHeader
 
 
-def flatten_column_header(column_header: Any):
+def flatten_column_header(column_header: ColumnHeader) -> ColumnHeader:
     """
     Given a pandas column header, if it is a list or tuple, it will
     flatten this header into a string. 
@@ -37,7 +45,7 @@ def flatten_column_header(column_header: Any):
 
     return column_header
 
-def get_column_header_display(column_header: Any) -> str:
+def get_column_header_display(column_header: ColumnHeader) -> str:
     """
     Turns the column header into a string that is the same as how
     the string is displayed on the front-end. 
@@ -56,12 +64,12 @@ def get_column_header_display(column_header: Any) -> str:
     return str(column_header)
 
 
-def get_column_header_ids(column_headers: List[Any]):
+def get_column_header_ids(column_headers: List[ColumnHeader]) -> List[ColumnID]:
     return [
         get_column_header_id(column_header) for column_header in column_headers
     ]
 
-def get_column_header_id(column_header: Any, use_deprecated_id_algorithm=False) -> str:
+def get_column_header_id(column_header: ColumnHeader, use_deprecated_id_algorithm: bool=False) -> ColumnID:
     """
     Returns the ID that represents this column header, which is
     used to refer to the column as long as it is used in the 
@@ -84,7 +92,7 @@ def get_column_header_id(column_header: Any, use_deprecated_id_algorithm=False)
     else:
         return str(column_header)
 
-def try_make_new_header_valid_if_multi_index_headers(column_headers: List[Any], new_column_header: Any):
+def try_make_new_header_valid_if_multi_index_headers(column_headers: List[ColumnHeader], new_column_header: ColumnHeader) -> ColumnHeader:
     """
     Helper function for when you're adding a column to a
     dataframe that has multi-index headers. 
@@ -93,11 +101,10 @@ def try_make_new_header_valid_if_multi_index_headers(column_headers: List[Any],
     that is the length of the multi-index headers, so
     this function helps you make such a header
     """
-    if not isinstance(new_column_header, tuple) and len(column_headers) > 0:
+    if (not isinstance(new_column_header, tuple) and not isinstance(new_column_header, list)) and len(column_headers) > 0:
         other_column_header = random.choice(list(column_headers))
         if isinstance(other_column_header, tuple):
-            ending_list = ['' for _ in range(len(other_column_header) - 1)]
-            ending_list.insert(0, new_column_header)
+            ending_list: MultiLevelColumnHeader = [new_column_header] + ['' for _ in range(len(other_column_header) - 1)]
             return tuple(ending_list)
 
     # If we don't need to change the column header, don't change it
@@ -112,9 +119,9 @@ class ColumnIDMap():
     two of them.
     """
 
-    def __init__(self, dfs: List[pd.DataFrame]):
-        self.column_id_to_column_header: List[Dict[str, Any]] = [dict() for _ in range(len(dfs))]
-        self.column_header_to_column_id: List[Dict[Any, str]] = [dict() for _ in range(len(dfs))]
+    def __init__(self, dfs: Collection[pd.DataFrame]):
+        self.column_id_to_column_header: List[Dict[ColumnID, ColumnHeader]] = [dict() for _ in range(len(dfs))]
+        self.column_header_to_column_id: List[Dict[ColumnHeader, ColumnID]] = [dict() for _ in range(len(dfs))]
 
         for sheet_index, df in enumerate(dfs):
             for column_header in df.keys():
@@ -122,7 +129,7 @@ class ColumnIDMap():
                 self.column_id_to_column_header[sheet_index][column_id] = column_header
                 self.column_header_to_column_id[sheet_index][column_header] = column_id
 
-    def set_column_header(self, sheet_index: int, column_id: str, column_header: Any):
+    def set_column_header(self, sheet_index: int, column_id: ColumnID, column_header: ColumnHeader) -> None:
         """
         Sets a column id and column header to match to eachother. 
 
@@ -130,7 +137,6 @@ class ColumnIDMap():
         headers, in which case even if the new column header is just a string, the
         the column header that ends up getting created in the dataframe will be a 
         tuple.
-        
         """
         # Make sure the new column header we're setting is in fact valid for this dataframe
         # header format - as it may be a multi-index header
@@ -145,7 +151,7 @@ class ColumnIDMap():
         self.column_id_to_column_header[sheet_index][column_id] = column_header
         self.column_header_to_column_id[sheet_index][column_header] = column_id
 
-    def add_df(self, df: pd.DataFrame, sheet_index: int=None, use_deprecated_id_algorithm=False) -> Dict[str, Any]:
+    def add_df(self, df: pd.DataFrame, sheet_index: int=None, use_deprecated_id_algorithm: bool=False) -> Dict[str, ColumnHeader]:
         """
         Adds all of the keys for the new dataframe to the column id 
         mappings.
@@ -169,7 +175,7 @@ class ColumnIDMap():
         
         return self.column_id_to_column_header[sheet_index]
 
-    def remove_df(self, sheet_index: int):
+    def remove_df(self, sheet_index: int) -> None:
         """
         Deletes the tracking of this dataframe from the column
         id map.
@@ -177,7 +183,7 @@ class ColumnIDMap():
         self.column_id_to_column_header.pop(sheet_index)
         self.column_header_to_column_id.pop(sheet_index)
     
-    def add_column_header(self, sheet_index: int, column_header: Any):
+    def add_column_header(self, sheet_index: int, column_header: ColumnHeader) -> str:
         """
         NOTE: this should only be called when adding a column to the dataframe,
         and not when renaming a column, as it creates a new id for the column
@@ -196,12 +202,12 @@ class ColumnIDMap():
 
         return column_id
     
-    def delete_column_id(self, sheet_index: int, column_id: str):
+    def delete_column_id(self, sheet_index: int, column_id: ColumnID) -> None:
         column_header = self.get_column_header_by_id(sheet_index, column_id)
         del self.column_id_to_column_header[sheet_index][column_id]
         self.column_header_to_column_id[sheet_index][column_header]
 
-    def get_column_ids(self, sheet_index: int, column_headers=None):
+    def get_column_ids(self, sheet_index: int, column_headers: Collection[ColumnHeader]=None) -> List[str]:
         if column_headers is None:
             return list(self.column_id_to_column_header[sheet_index].keys())
         try:
@@ -212,28 +218,38 @@ class ColumnIDMap():
         except:
             raise make_no_column_error(column_headers)
 
-    def get_column_ids_map(self, sheet_index: int):
+    def get_column_headers(self, sheet_index: int) -> List[ColumnHeader]:
+        return list(self.column_id_to_column_header[sheet_index].values())
+
+    def get_column_ids_map(self, sheet_index: int) -> Dict[str, ColumnHeader]:
         return self.column_id_to_column_header[sheet_index]
 
-    def get_column_id_by_header(self, sheet_index: int, column_header: Any):
+    def get_column_id_by_header(self, sheet_index: int, column_header: ColumnHeader) -> str:
         return self.column_header_to_column_id[sheet_index][column_header]
 
-    def get_column_header_by_id(self, sheet_index: int, column_id: str):
+    def get_column_ids_by_headers(self, sheet_index: int, column_headers: List[ColumnHeader]) -> List[ColumnID]:
+        return [
+            self.column_header_to_column_id[sheet_index][column_header] 
+            for column_header in column_headers
+        ]
+
+    def get_column_header_by_id(self, sheet_index: int, column_id: ColumnID) -> ColumnHeader:
         return self.column_id_to_column_header[sheet_index][column_id]
 
-    def get_column_headers_by_ids(self, sheet_index: int, column_ids: List[str]):
+    def get_column_headers_by_ids(self, sheet_index: int, column_ids: List[ColumnID]) -> List[ColumnHeader]:
         return [
             self.column_id_to_column_header[sheet_index][column_id] 
             for column_id in column_ids
         ]
 
-    def move_to_deprecated_id_format(self):
+    def move_to_deprecated_id_format(self) -> None:
         """
         Helper function to change the column IDs, specifically for the 
         purpose of backwards compability. This simulates a preprocessing 
         step.
         """
-        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import make_valid_header
+        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import \
+            make_valid_header
         for sheet_index in range(len(self.column_id_to_column_header)):
             column_id_to_column_header_map = self.column_id_to_column_header[sheet_index]
 
